"""Functions that collects metadata"""

import json
import logging
from pathlib import Path

from crunchy.integrity import get_checksum

LOG = logging.getLogger(__name__)


def get_fastq_info(fastq: Path, tag: str, algorithm) -> dict:
    """Get the necessary information about a fastq file and return it in a dict"""
    fastq_info = {"file": tag}
    fastq_info["checksum"] = get_checksum(infile=fastq, algorithm=algorithm)
    fastq_info["path"] = str(fastq.absolute())
    fastq_info["algorithm"] = algorithm

    return fastq_info


def fetch_spring_metadata(
    first_read: Path, second_read: Path, spring: Path, algorithm: str = "sha256"
) -> list:
    """Create metadata for a spring archive

    This means that for each file the original paths and checksums are stored in a dict.
    """
    metadata = []
    metadata.append(get_fastq_info(first_read, "first_read", algorithm))
    metadata.append(get_fastq_info(second_read, "second_read", algorithm))
    metadata.append({"path": str(spring.absolute()), "file": "spring"})

    return metadata


def dump_spring_metadata(metadata: list) -> Path:
    """Write spring metadata to json file

    Path to metadata is the same as the path to spring archive with suffix .json instead of .spring

    Returns:
        metadata_path
    """
    spring_path = None
    for file_info in metadata:
        if file_info["file"] == "spring":
            spring_path = Path(file_info["path"])

    metadata_path = spring_path.with_suffix(".json")
    with open(metadata_path, "w") as out:
        LOG.info("Dumping spring metadata to %s", metadata_path)
        json.dump(metadata, out, indent=2)
    return metadata_path



"""Tests for compare CLI"""

from click.testing import CliRunner

from crunchy.cli.base import base_command
from crunchy.integrity import get_checksum


def test_compare_same(first_read):
    """Test to compare a file with itself"""
    # GIVEN the path to a gzipped file and a cli runner
    runner = CliRunner()
    # WHEN running the compare command
    result = runner.invoke(base_command, ["compare", "-f", str(first_read), "-s", str(first_read)])
    # THEN assert the command was succesful
    assert result.exit_code == 0


def test_compare_checksum(first_read):
    """Test to compare a file with a checksum"""
    # GIVEN the path to a gzipped file, a checksum and a cli runner
    checksum = get_checksum(first_read)
    runner = CliRunner()
    # WHEN running the compare command
    result = runner.invoke(base_command, ["compare", "-f", str(first_read), "-c", checksum])
    # THEN assert the command was succesful
    assert result.exit_code == 0


def test_compare_checksum_two_reads(first_read):
    """Test to compare two files AND a checksum"""
    # GIVEN the path to a gzipped file, a checksum and a cli runner
    checksum = get_checksum(first_read)
    runner = CliRunner()
    # WHEN running the compare command with both files and a checksum
    result = runner.invoke(
        base_command,
        ["compare", "-f", str(first_read), "-s", str(first_read), "-c", checksum],
    )
    # THEN assert the command fails since two files can not be compared to one checksum
    assert result.exit_code == 1


def test_compare_wrong_checksum(first_read, second_read):
    """Test to compare a file with a wrong checksum"""
    # GIVEN the path to a gzipped file and a cli runner
    runner = CliRunner()
    # GIVEN a mismatching checksum
    checksum = get_checksum(second_read)
    # WHEN running the compare command with one file and the wrong checksum
    result = runner.invoke(
        base_command,
        ["compare", "-f", str(first_read), "-c", checksum],
    )
    # THEN assert the command fails checksums differ
    assert result.exit_code == 1


def test_compare_different(first_read, second_read):
    """Test to compare two different files"""
    # GIVEN the path to a gzipped file and a cli runner
    runner = CliRunner()
    # WHEN running the compare command
    result = runner.invoke(base_command, ["compare", "-f", str(first_read), "-s", str(second_read)])
    # THEN assert the command was succesful
    assert result.exit_code == 1


def test_check_logging(first_read):
    """Test to compare a file with itself"""
    # GIVEN the path to a gzipped file and a cli runner
    runner = CliRunner()
    # GIVEN a checksum for the read
    checksum = get_checksum(first_read)
    # WHEN running the compare command
    result = runner.invoke(base_command, ["compare", "-f", str(first_read), "-s", str(first_read)])
    # THEN assert the checksum that is logged is the same
    new_checksum = None
    for message in result.output.split("\n"):
        if "Checksum:" in message:
            new_checksum = message.split(" ")[-1]

    assert checksum == new_checksum



"""Code to test the CLI compress commands"""

import logging
import pathlib

from click.testing import CliRunner

from crunchy.cli.compress_cmd import bam, compress, fastq

LOG = logging.getLogger(__name__)


def test_compress_cmd():
    """Test to run the compress base command"""
    # GIVEN a cli runner
    runner = CliRunner()
    # WHEN running the compress command with dry_run
    result = runner.invoke(compress, obj={})
    # THEN assert the command was succesful even without a valid api
    assert result.exit_code == 0


def test_compress_bam_dry_run(bam_tmp_file, base_context):
    """Test to run the compress bam command"""
    # GIVEN the path to a existing bam file and a cli runner
    runner = CliRunner()
    bam_path = bam_tmp_file
    assert bam_path.exists()
    # WHEN running the compress command with dry_run
    result = runner.invoke(bam, ["--bam-path", str(bam_path), "--dry-run"], obj=base_context)
    # THEN assert the command was succesful
    assert result.exit_code == 0


def test_compress_bam_no_outpath(base_context, bam_tmp_file):
    """Test to run the compress bam command"""
    # GIVEN the path to a bam file and a cli runner
    runner = CliRunner()
    bam_path = bam_tmp_file
    assert bam_path.exists()
    # WHEN running the compress command
    result = runner.invoke(
        bam,
        ["--bam-path", bam_path],
        obj=base_context,
    )
    # THEN assert the command was succesful
    assert result.exit_code == 0


def test_compress_bam_valid_outpath(base_context, bam_path):
    """Test to run the compress bam command"""
    # GIVEN the path to a bam file, a non existing outpath and a cli runner
    outpath = "a_file.cram"
    runner = CliRunner()
    # WHEN running the compress command
    result = runner.invoke(
        bam,
        ["--bam-path", str(bam_path), "--cram-path", outpath],
        obj=base_context,
    )
    # THEN assert the command was succesful
    assert result.exit_code == 0


def test_compress_bam_existing_outpath(base_context, bam_path, cram_path):
    """Test to run the compress bam command"""
    # GIVEN the path to a bam file, a existing outpath and a cli runner
    runner = CliRunner()
    assert cram_path.exists()
    # WHEN running the compress command
    res = runner.invoke(
        bam,
        ["--bam-path", str(bam_path), "--cram-path", str(cram_path)],
        obj=base_context,
    )
    # THEN the progam should abort since the cram file already exists
    assert res.exit_code == 1


def test_compress_fastq_dry_run(first_read, second_read):
    """Test to run the compress fastq command"""
    # GIVEN the path to a existing bam file and a cli runner
    runner = CliRunner()
    assert first_read.exists()
    assert second_read.exists()
    # WHEN running the compress command with dry_run
    result = runner.invoke(
        fastq,
        [
            "--first-read",
            str(first_read),
            "--second-read",
            str(second_read),
            "--dry-run",
        ],
        obj={},
    )
    # THEN assert the command was succesful even without a valid api
    assert result.exit_code == 0


def test_compress_fastq_dry_run_integrity(first_read, second_read):
    """Test to run the compress fastq with integrity check in dry run mode"""
    # GIVEN the path to a existing bam file and a cli runner
    runner = CliRunner()
    assert first_read.exists()
    assert second_read.exists()
    # WHEN running the compress command with dry_run
    result = runner.invoke(
        fastq,
        [
            "--first-read",
            str(first_read),
            "--second-read",
            str(second_read),
            "--dry-run",
            "--check-integrity",
        ],
        obj={},
    )
    # THEN assert the command was succesful even without a valid api
    assert result.exit_code == 0


def test_compress_fastq_existing_spring_file(first_read, second_read, spring_path, base_context):
    """Test to run the compress fastq command"""
    # GIVEN the path to a existing bam file and a cli runner
    runner = CliRunner()
    assert spring_path.exists()
    # WHEN running the compress command with dry_run
    result = runner.invoke(
        fastq,
        [
            "--first-read",
            str(first_read),
            "--second-read",
            str(second_read),
            "--spring-path",
            str(spring_path),
        ],
        obj=base_context,
    )
    # THEN assert the command failed since the spring file exists
    assert result.exit_code == 1


def test_compress_fastq_valid_spring_file(first_read, second_read, spring_tmp_path, base_context):
    """Test to run the compress fastq command"""
    # GIVEN the path to a existing bam file and a cli runner
    runner = CliRunner()
    assert not spring_tmp_path.exists()
    # WHEN running the compress command with dry_run
    result = runner.invoke(
        fastq,
        [
            "--first-read",
            str(first_read),
            "--second-read",
            str(second_read),
            "--spring-path",
            str(spring_tmp_path),
        ],
        obj=base_context,
    )
    # THEN assert the command succedes
    assert result.exit_code == 0


def test_compress_fastq_with_metadata(
    first_read, second_read, spring_tmp_path, base_context, metadata_tmp_path
):
    """Test to run the compress fastq command with metadata written"""
    # GIVEN the path a pair of fastqs, a spring file and a cli runner
    runner = CliRunner()
    # GIVEN a non existing spring path
    assert not spring_tmp_path.exists()
    # GIVEN a non existing metadata path
    assert not metadata_tmp_path.exists()
    # WHEN running the compress command with metadata
    result = runner.invoke(
        fastq,
        [
            "--first-read",
            str(first_read),
            "--second-read",
            str(second_read),
            "--spring-path",
            str(spring_tmp_path),
            "--metadata-file",
        ],
        obj=base_context,
    )
    # THEN assert the command succedes
    assert result.exit_code == 0
    # THEN assert the metadata file was created
    assert metadata_tmp_path.exists()


def nr_files(dirpath: pathlib.Path) -> int:
    """Count the number of files in a directory"""
    nr_files_indir = 0
    for path in dirpath.iterdir():
        if not path.is_file():
            continue
        nr_files_indir += 1
    return nr_files_indir



"""Utilities for crunchy cli functions"""

import logging
import pathlib

import click

LOG = logging.getLogger(__name__)


def file_exists(file_path: pathlib.Path, exists: bool = True) -> bool:
    """Check if a file exists.

    If not raise a click.Abort("File <file_path> does not exist")
    If make sure that file does not exist set exists to False
    """
    if file_path.exists():
        if exists is False:
            LOG.error("File %s already exists!", file_path)
            raise click.Abort
    else:
        if exists is True:
            LOG.error("Could not find file %s", file_path)
            raise click.Abort

    return True



"""Base conftest file."""

import logging
import shutil
import sys
from pathlib import Path
from typing import Generator

import pytest

from crunchy.command import CramProcess, SpringProcess
from crunchy.integrity import get_checksum

LOG = logging.getLogger(__name__)
logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)


@pytest.fixture(name="fixtures_dir")
def fixture_fixtures_dir() -> Path:
    """Return the path to the fixture's dir."""
    return Path("tests/fixtures")


# File paths fixtures #


@pytest.fixture(name="reference_path")
def fixture_reference_path(fixtures_dir: Path) -> Path:
    """Return the path to FASTA reference."""
    return Path(fixtures_dir, "reference.fasta")


@pytest.fixture
def dummy_file_path(fixtures_dir: Path) -> Path:
    """Return the path to a dummy file."""
    return Path(fixtures_dir, "dummy.txt")


@pytest.fixture
def zipped_file_path(fixtures_dir: Path) -> Path:
    """Return the path to a zipped dummy file."""
    return Path(fixtures_dir, "zipped_file.txt.gz")


@pytest.fixture(name="bam_path")
def fixture_bam_path(fixtures_dir: Path) -> Path:
    """Return the path to a BAM file."""
    return Path(fixtures_dir, "bam", "test.bam")


@pytest.fixture(name="cram_path")
def fixture_cram_path(fixtures_dir: Path) -> Path:
    """Return the path to a CRAN file"""
    return Path(fixtures_dir, "bam", "test.cram")


@pytest.fixture(name="first_read")
def fixture_first_read(fixtures_dir: Path) -> Path:
    """Return the path first read in read pair."""
    return Path(fixtures_dir, "fastq", "TEST_R1_001.fq.gz")


@pytest.fixture(name="second_read")
def fixture_second_read(fixtures_dir: Path) -> Path:
    """Return the path second read in read pair."""
    return Path(fixtures_dir, "fastq", "TEST_R2_001.fq.gz")


@pytest.fixture(name="spring_path")
def fixture_spring_path(fixtures_dir: Path) -> Path:
    """Return the path to a Spring compressed file."""
    return Path(fixtures_dir, "spring", "TEST.spring")


# Temp files fixtures #


@pytest.fixture(scope="function", name="project_dir")
def fixture_project_dir(tmpdir_factory) -> Generator[Path, None, None]:
    """Path to a temporary directory."""
    yield Path(tmpdir_factory.mktemp("data"))


# Paths to non existing files #


@pytest.fixture(name="first_tmp_path")
def fixture_first_tmp_path(project_dir: Path, first_read: Path) -> Path:
    """Return the path to a nonexistent FASTQ file."""
    return Path(project_dir, first_read.name)


@pytest.fixture(name="second_tmp_path")
def fixture_second_tmp_path(project_dir: Path, second_read: Path) -> Path:
    """Return the path to a nonexistent FASTQ file."""
    return Path(project_dir, second_read.name)


@pytest.fixture(name="spring_tmp_path")
def fixture_spring_tmp_path(project_dir: Path, spring_path: Path) -> Path:
    """Return the path to a nonexistent temporary Spring file."""
    return Path(project_dir, spring_path.name)


@pytest.fixture(name="bam_tmp_path")
def fixture_bam_tmp_path(bam_path: Path, project_dir: Path) -> Path:
    """Return the path to a nonexistent small BAM file."""
    return Path(project_dir, bam_path.name)


@pytest.fixture(name="cram_tmp_path")
def fixture_cram_tmp_path(cram_path: Path, project_dir: Path) -> Path:
    """Return the path to a nonexistent temporary CRAM file."""
    return Path(project_dir, cram_path.name)


@pytest.fixture(name="cram_tmp_index_path")
def fixture_cram_tmp_index_path(cram_tmp_path: Path, real_cram_api: CramProcess) -> Path:
    """Return the path to a nonexistent temporary CRAM file."""
    return real_cram_api.get_index_path(cram_tmp_path)


@pytest.fixture(name="metadata_tmp_path")
def fixture_metadata_tmp_path(spring_tmp_path: Path) -> Path:
    """Return the path to a nonexisting temporary spring metadata file."""
    return spring_tmp_path.with_suffix(".json")


# Paths to existing temporary files #


@pytest.fixture(name="first_tmp_file")
def fixture_first_tmp_file(first_tmp_path: Path, first_read: Path) -> Path:
    """Return the path to a temporary FASTQ file."""
    shutil.copy(first_read.as_posix(), first_tmp_path.as_posix())
    return first_tmp_path


@pytest.fixture(name="second_tmp_file")
def fixture_second_tmp_file(second_tmp_path: Path, second_read: Path) -> Path:
    """Return the path to a temporary FASTQ file."""
    shutil.copy(second_read.as_posix(), second_tmp_path.as_posix())
    return second_tmp_path


@pytest.fixture(name="spring_tmp_file")
def fixture_spring_tmp_file(spring_tmp_path: Path, spring_path: Path) -> Path:
    """Return the path to a temporary Spring file."""
    shutil.copy(spring_path.as_posix(), spring_tmp_path.as_posix())
    return spring_tmp_path


@pytest.fixture
def bam_tmp_file(bam_path: Path, bam_tmp_path: Path) -> Path:
    """Return the path to a temporary small bam file"""
    shutil.copy(bam_path.as_posix(), bam_tmp_path.as_posix())
    return bam_tmp_path


@pytest.fixture
def cram_tmp_file(cram_path: Path, cram_tmp_path: Path) -> Path:
    """Return the path to a temporary CRAM file."""
    shutil.copy(cram_path.as_posix(), cram_tmp_path.as_posix())
    return cram_tmp_path


# Fixtures for checksums #


@pytest.fixture(name="checksum_first_read")
def fixture_checksum_first_read(first_read: Path) -> str:
    """Return the checksum for first FASTQ read."""
    return get_checksum(first_read)


@pytest.fixture(name="checksum_second_read")
def fixture_checksum_second_read(second_read: Path) -> str:
    """Return the checksum for second FASTQ read."""
    return get_checksum(second_read)


# Fixtures for metadata


@pytest.fixture(name="spring_metadata")
def fixture_spring_metadata(
    first_read: Path,
    second_read: Path,
    spring_tmp_path: Path,
    checksum_first_read: str,
    checksum_second_read: str,
):
    """Return metadata information."""
    return [
        {
            "path": str(first_read.absolute()),
            "file": "first_read",
            "checksum": checksum_first_read,
            "algorithm": "sha256",
        },
        {
            "path": str(second_read.absolute()),
            "file": "second_read",
            "checksum": checksum_second_read,
            "algorithm": "sha256",
        },
        {"path": str(spring_tmp_path.absolute()), "file": "spring"},
    ]


class MockSpringProcess:
    """Mock the Spring API."""

    def __init__(self, binary="spring", threads=8, tmp_dir=None):
        """Initialize a MockSpringProcess."""
        self.binary = binary
        self.threads = threads
        self.base_call = [self.binary]
        self.tmp = tmp_dir or "./tmp"
        self._create_output = False
        self._fastq1 = None
        self._fastq2 = None

    @staticmethod
    def run_command(parameters=None):
        """Mock out the run command functionality."""
        LOG.info("Running command %s", " ".join(parameters))
        return 0

    def decompress(self, spring_path: Path, first: Path, second: Path) -> bool:
        """Run the spring decompress command."""
        parameters = ["-d", "-i", spring_path.as_posix(), "-o", first.as_posix(), second.as_posix()]
        self.run_command(parameters)
        if self._create_output:
            LOG.info(f"Create output fastq files {self._fastq1} and {self._fastq2}")
            shutil.copy(str(self._fastq1), first.as_posix())
            shutil.copy(str(self._fastq2), second.as_posix())
        return True

    def compress(self, first: Path, second: Path, outfile: Path) -> bool:
        """Run the spring compression command."""
        parameters = [
            "-c",
            "-i",
            first.as_posix(),
            second.as_posix(),
            "-o",
            outfile.as_posix(),
            "-t",
            str(self.threads),
        ]
        self.run_command(parameters)
        return True


class MockCramProcess:
    """Mock the CRAM API."""

    def __init__(self, binary="samtools", refgenome_path="genome.fasta", threads=8):
        """Initialize a CramProcessMock."""
        self.binary = binary
        self.refgenome_path = refgenome_path
        self.threads = threads
        self.base_call = [self.binary]

    @staticmethod
    def run_command(parameters=None) -> int:
        """Execute a command in the shell.

        Args:
            parameters(list)
        """
        LOG.info("Running command %s", " ".join(parameters))
        return 0

    def decompress(self, cram_path: Path, bam_path: Path) -> bool:
        """Convert CRAM to BAM."""
        LOG.info("Decompressing cram %s to bam %s", cram_path, bam_path)
        parameters = [
            "view",
            "-b",
            "-o",
            bam_path.as_posix(),
            "-r",
            self.refgenome_path,
            cram_path.as_posix(),
        ]
        self.run_command(parameters)
        return True

    def compress(self, bam_path: Path, cram_path: Path) -> bool:
        """Convert BAM to CRAM."""
        LOG.info(f"Compressing bam {bam_path} to cram {cram_path}")
        parameters = [
            "view",
            "-C",
            "-T",
            self.refgenome_path,
            bam_path.as_posix(),
            "-o",
            cram_path.as_posix(),
        ]
        self.run_command(parameters)
        self.index(cram_path)
        return True

    def index(self, file_path: Path) -> bool:
        """Index a BAN or CRAM file."""
        LOG.info("Creating index for %s", file_path)
        index_type: str = ".bai" if file_path.suffix == ".bam" else ".cram"
        index_path: Path = file_path.with_suffix(file_path.suffix + index_type)
        parameters: list[str] = ["index", file_path.as_posix(), index_path.as_posix()]
        self.run_command(parameters)
        return True

    @staticmethod
    def self_check():
        """Mocks the self test."""
        return True


# Fixtures for apis #


@pytest.fixture(name="spring_api")
def fixture_spring_api() -> MockSpringProcess:
    """Return a mocked Spring API."""
    return MockSpringProcess("spring", threads=8)


@pytest.fixture(name="cram_api")
def fixture_cram_api() -> MockCramProcess:
    """Return a mocked Spring API."""
    return MockCramProcess("samtools", threads=8)


@pytest.fixture(name="real_spring_api")
def fixture_real_spring_api() -> SpringProcess:
    """Return a Spring API that runs Spring"""
    return SpringProcess("spring", threads=8)


@pytest.fixture(name="real_cram_api")
def fixture_real_cram_api(reference_path: Path) -> CramProcess:
    """Return a CRAM API."""
    return CramProcess("samtools", refgenome_path=str(reference_path))



"""Fixtures for CLI tests"""

import tempfile
from pathlib import Path
from typing import Any, Generator

import pytest

from crunchy.command import CramProcess, SpringProcess
from tests.conftest import MockCramProcess, MockSpringProcess


@pytest.fixture(name="base_context")
def fixture_base_context(
    spring_api: MockSpringProcess, cram_api: MockCramProcess
) -> dict[str, Any]:
    """Return base context to use in CLI."""
    return {
        "spring_api": spring_api,
        "cram_api": cram_api,
    }


@pytest.fixture(name="real_base_context")
def fixture_real_base_context(
    real_spring_api: SpringProcess, real_cram_api: CramProcess
) -> dict[str, Any]:
    """Return real base context to use in CLI."""
    return {
        "spring_api": real_spring_api,
        "cram_api": real_cram_api,
    }


@pytest.fixture(name="non_existing_path")
def fixture_non_existing_path() -> Generator[Path, None, None]:
    """Return the path tp a non-existing file."""
    yield Path(tempfile.NamedTemporaryFile().name)



"""Code for checksum cli command"""

import logging
import pathlib

import click

from crunchy.integrity import get_checksum

LOG = logging.getLogger(__name__)


@click.command()
@click.argument("infile", type=click.Path(exists=True))
@click.option("--algorithm", "-a", type=click.Choice(["md5", "sha1", "sha256"]), default="sha256")
def checksum(infile, algorithm):
    """Generate the checksum for a file."""
    LOG.info("Running checksum")
    click.echo(get_checksum(pathlib.Path(infile), algorithm))



"""Code to test crunchys files module"""

import pathlib

from crunchy import files


def test_cram_output():
    """Test to generate a cram path from bam"""
    # GIVEN a bam path
    abam = pathlib.Path("test.bam")
    # WHEN fetching the cram path
    acram = files.cram_outpath(abam)
    # THEN assert the filename is the same
    assert abam.with_suffix("") == acram.with_suffix("")
    assert str(abam) != str(acram)



"""Code to handle integrity of files."""

import gzip
import hashlib
import logging
from pathlib import Path

from typing import Any

LOG = logging.getLogger(__name__)


def compare_elements(elements: list) -> bool:
    """Check if all elements are the same."""
    return len(set(elements)) == 1


def get_checksum(infile: Path, algorithm: str = "sha256") -> str:
    """Get the checksum for a file."""
    LOG.info(f"Create checksum for {infile}")
    if algorithm == "sha1":
        LOG.info("Use sha1")
        hash_obj = hashlib.sha1()
    elif algorithm == "md5":
        LOG.info("Use md5")
        hash_obj = hashlib.md5()
    else:
        LOG.info("Use sha256")
        hash_obj = hashlib.sha256()

    content: Any = open(infile, "rb")
    if infile.suffix in [".gz", ".gzip"]:
        LOG.info("Unzip before counting checksum")
        content = gzip.open(infile, "rb")

    return generate_checksum(content, hash_obj)


def generate_checksum(content: Any, hash_obj: "hashlib._HASH") -> str:
    """Return the checksum of a file.

    Args:
        content(iterable)
        hash_obj
    """
    for chunk in iter(lambda: content.read(4096), b""):
        hash_obj.update(chunk)
    LOG.info("Checksum created")
    return hash_obj.hexdigest()



"""Tests for utils"""

from crunchy.files import spring_outpath
from crunchy.utils import find_fastq_pairs


def test_find_fastq_pairs_fixtures(fixtures_dir, first_read, second_read):
    """Test to find all fastq pairs in a directory"""
    # GIVEN the path to a directory with fastq files
    assert fixtures_dir.is_dir()
    spring_path = spring_outpath(first_read)
    # WHEN finding all fastq pairs
    pairs = find_fastq_pairs(fixtures_dir)
    res = next(pairs)
    # THEN assert that the correct reads are retuned
    assert res[0] == first_read
    assert res[1] == second_read
    assert res[2] == spring_path



"""Code for dealing with fastq paths"""

import logging
import pathlib

LOG = logging.getLogger(__name__)


def cram_outpath(bam_path: pathlib.Path) -> pathlib.Path:
    """Build a bam path based on a cram path"""
    LOG.info("Create cram path from %s", bam_path)
    file_name = bam_path.with_suffix("")
    return file_name.with_suffix(".cram")


def fastq_outpaths(filepath: pathlib.Path) -> tuple:
    """Build a pair of fastq paths based on a file path"""
    LOG.info("Create fastq paths from %s", filepath)
    file_name = filepath.name
    file_parent = filepath.parent
    pair_str = str(pathlib.Path(file_name).with_suffix("")) + "_R{}_" + "001.fastq.gz"
    return (
        pathlib.Path(file_parent, pair_str.format("1")),
        pathlib.Path(file_parent, pair_str.format("2")),
    )


def spring_outpath(filepath: pathlib.Path) -> pathlib.Path:
    """Build a spring path based on a fastq file path"""
    LOG.info("Create spring path from %s", filepath)
    file_name = filepath.name
    file_parent = filepath.parent

    splitted = file_name.split("_")
    spring_base = pathlib.Path("_".join(splitted[:-2]))

    spring_path = pathlib.Path(file_parent).joinpath(spring_base).with_suffix(".spring")
    LOG.info("Creates spring path %s", spring_path)

    return spring_path



"""Code to compress a pair of fastq files"""

import logging
import pathlib

from .command import CramProcess, SpringProcess

LOG = logging.getLogger(__name__)


def compress_spring(
    first_read: pathlib.Path,
    second_read: pathlib.Path,
    spring_api: SpringProcess,
    outfile: pathlib.Path,
    dry_run: bool = False,
) -> bool:
    """Compress file(s)"""
    first_read = first_read.absolute()
    second_read = second_read.absolute()
    outfile = outfile.absolute()
    LOG.info("Compressing %s and %s to %s", first_read, second_read, outfile)
    if dry_run:
        return True

    return spring_api.compress(first=first_read, second=second_read, outfile=outfile)


def compress_cram(
    bam_path: pathlib.Path,
    cram_path: pathlib.Path,
    cram_api: CramProcess,
    dry_run: bool = False,
) -> bool:
    """Compress bam file"""
    bam_path = bam_path.absolute()
    cram_path = cram_path.absolute()

    LOG.info("Compressing %s to %s", bam_path, cram_path)
    if dry_run:
        return True

    return cram_api.compress(bam_path=bam_path, cram_path=cram_path)



"""Tests for decompress functions."""

from pathlib import Path

from crunchy.decompress import decompress_cram, decompress_spring
from crunchy.files import spring_outpath
from tests.conftest import MockCramProcess, MockSpringProcess


def test_decompress_spring(first_read: Path, second_read: Path, spring_api: MockSpringProcess):
    """Test top decompress compressed file."""
    # GIVEN a compressed file and a outfile
    spring_path = spring_outpath(first_read)

    # WHEN decompressing
    res = decompress_spring(
        spring_path=spring_path, first=first_read, second=second_read, spring_api=spring_api
    )

    # THEN assert that the function returns True
    assert res is True


def test_decompress_spring_dry_run(
    first_read: Path, second_read: Path, spring_api: MockSpringProcess
):
    """Test top decompress compressed file."""
    # GIVEN a compressed file and a outfile
    spring_path = spring_outpath(first_read)

    # WHEN decompressing
    res = decompress_spring(
        spring_path=spring_path,
        first=first_read,
        second=second_read,
        spring_api=spring_api,
        dry_run=True,
    )

    # THEN assert that the function returns True even if no API is used
    assert res is True


def test_decompress_cram(cram_path: Path, bam_path: Path, cram_api: MockCramProcess):
    """Test to decompress a cram file,"""
    # GIVEN a cram file, a bam file and a cram api

    # WHEN decompressing
    res = decompress_cram(cram_path=cram_path, bam_path=bam_path, cram_api=cram_api)

    # THEN assert that the process return True
    assert res is True


def test_decompress_cram_dry_run(cram_path: Path, bam_path: Path, cram_api: MockCramProcess):
    """Test to decompress a cram file."""
    # GIVEN a cram file, a bam file and a cram api

    # WHEN decompressing in dry run mode
    res = decompress_cram(cram_path=cram_path, bam_path=bam_path, cram_api=cram_api, dry_run=True)

    # THEN assert that the process returns true
    assert res is True



"""CLI functions to compress."""

import logging
from pathlib import Path
from typing import Optional

import click

from crunchy.cli.compare_cmd import compare
from crunchy.cli.decompress_cmd import spring as decompress_spring_cmd
from crunchy.cli.utils import file_exists
from crunchy.compress import compress_cram, compress_spring
from crunchy.files import cram_outpath, spring_outpath
from crunchy.metadata import dump_spring_metadata, fetch_spring_metadata

LOG = logging.getLogger(__name__)


@click.group()
def compress():
    """Compress genomic files."""


@click.command()
@click.option(
    "--first-read",
    "-f",
    type=click.Path(exists=True),
    required=True,
    help="First read in pair",
)
@click.option(
    "--second-read",
    "-s",
    type=click.Path(exists=True),
    required=True,
    help="Second read in pair",
)
@click.option(
    "--spring-path",
    "-o",
    help="Path to spring file",
)
@click.option(
    "--check-integrity",
    is_flag=True,
    help="If the integrity of the files should be checked",
)
@click.option("--dry-run", is_flag=True)
@click.option(
    "--metadata-file",
    is_flag=True,
    help="If a json file with metada should be produced",
)
@click.pass_context
def fastq(ctx, first_read, second_read, spring_path, dry_run, check_integrity, metadata_file):
    """Compress a pair of FASTQ files with Spring."""
    LOG.info("Running compress fastq")
    if dry_run:
        LOG.warning("Dry Run! No files will be created or deleted")

    spring_api = ctx.obj.get("spring_api")
    first_read = Path(first_read)
    second_read = Path(second_read)
    spring_path = Path(spring_path) if spring_path else spring_outpath(first_read)

    file_exists(spring_path, exists=False)

    compress_spring(
        first_read=first_read,
        second_read=second_read,
        outfile=spring_path,
        spring_api=spring_api,
        dry_run=dry_run,
    )

    metadata = fetch_spring_metadata(
        first_read=first_read, second_read=second_read, spring=spring_path
    )

    metadata_path: Optional[Path] = dump_spring_metadata(metadata) if metadata_file else None
    if not check_integrity:
        return

    first_spring = Path(first_read).with_suffix(".spring.fastq")
    second_spring = Path(second_read).with_suffix(".spring.fastq")

    ctx.invoke(
        decompress_spring_cmd,
        spring_path=str(spring_path),
        first_read=str(first_spring),
        second_read=str(second_spring),
        dry_run=dry_run,
    )

    checksums = [None, None]
    for file_info in metadata:
        if file_info["file"] == "first_read":
            checksums[0] = file_info["checksum"]
        elif file_info["file"] == "second_read":
            checksums[1] = file_info["checksum"]

    success = True
    try:
        ctx.invoke(compare, first=str(first_spring), checksum=checksums[0], dry_run=dry_run)
        ctx.invoke(compare, first=str(second_spring), checksum=checksums[1], dry_run=dry_run)
    except click.Abort:
        LOG.error("Uncompressed Spring differ from original FASTQs")
        success = False
        LOG.info(f"Deleting compressed spring file {spring_path}")
        spring_path.unlink()
        if metadata_file:
            LOG.info(f"Deleting metadata file {metadata_path}")
            metadata_path.unlink()

    LOG.info("Deleting decompressed spring files")
    if not dry_run:
        first_spring.unlink()
        LOG.info(f"{first_spring} deleted")
        second_spring.unlink()
        LOG.info(f"{second_spring} deleted")

    if not success:
        raise click.Abort
    LOG.info("Files are identical, compression successful")


@click.command()
@click.option(
    "--bam-path",
    "-b",
    type=click.Path(exists=True),
    required=True,
    help="Path to bam file",
)
@click.option(
    "--cram-path",
    "-c",
    help="Path to cram file",
)
@click.option("--dry-run", is_flag=True)
@click.pass_context
def bam(ctx, bam_path: click.Path, cram_path: str, dry_run: bool):
    """Compress a BAM file to CRAM format with Samtools."""
    LOG.info("Running compress BAM")
    if dry_run:
        LOG.info("Dry Run! No files will be created or deleted")
    cram_api = ctx.obj.get("cram_api")
    try:
        cram_api.self_check()
    except (SyntaxError, FileNotFoundError) as error:
        raise click.Abort from error
    bam_path: Path = Path(bam_path)
    cram_path: Path = Path(cram_path) if cram_path else cram_outpath(bam_path)
    file_exists(cram_path, exists=False)
    compress_cram(
        bam_path=bam_path,
        cram_path=cram_path,
        cram_api=cram_api,
        dry_run=dry_run,
    )

    LOG.info("Compression successful")


for sub_cmd in [fastq, bam]:
    compress.add_command(sub_cmd)



"""Tests for metadata module"""

from crunchy import metadata


def test_fetch_spring_metadata(first_read, second_read, spring_path):
    """Test to fetch the metadata for a spring archive"""
    # GIVEN a pair of fastq files and a spring file

    # WHEN creating the metadata
    res = metadata.fetch_spring_metadata(
        first_read=first_read, second_read=second_read, spring=spring_path
    )

    # THEN assert that the result is a list
    assert isinstance(res, list)

    # THEN assert that there is one entry per file
    assert len(res) == 3


def test_get_fastq_info(first_read):
    """Test to create metadata info for a fastq file"""
    # GIVEN a fastq file

    # WHEN fetching the file info
    res = metadata.get_fastq_info(fastq=first_read, tag="first_read", algorithm="sha256")

    # THEN assert that the result is a dictionary
    assert isinstance(res, dict)

    # THEN assert that the checksum is there
    assert isinstance(res["checksum"], str)


def test_dump_metadata(spring_metadata, metadata_tmp_path):
    """Test to dump some metadata in json format"""
    # GIVEN some metadata and the path to a metadata file that does not exist
    assert not metadata_tmp_path.exists()

    # WHEN dumping the metadata
    metadata.dump_spring_metadata(spring_metadata)

    # THEN assert that the metadata file was created
    assert metadata_tmp_path.exists()



"""Code for CLI auto command"""

import logging
import pathlib

import click

from crunchy.utils import find_fastq_pairs

from .compress_cmd import fastq as compress_fastq_cmd

LOG = logging.getLogger(__name__)


def abort_if_false(ctx, param, value):
    """Check if abort"""
    if not value:
        ctx.abort()


@click.group()
def auto():
    """Run whole pipeline by compressing, comparing and deleting original files."""
    LOG.info("Running crunchy auto")


@click.command()
@click.option("--indir", type=click.Path(exists=True))
@click.option("--dry-run", is_flag=True)
@click.option("--spring-path", type=click.Path(exists=False))
@click.option(
    "--first",
    "-f",
    type=click.Path(exists=True),
    help="First in fastq pair to compare",
)
@click.option(
    "--second",
    "-s",
    type=click.Path(exists=True),
    help="Second in fastq pair to compare",
)
@click.option(
    "--yes",
    is_flag=True,
    callback=abort_if_false,
    expose_value=False,
    prompt="Are you sure you want to compress and delete all fastqs",
)
@click.pass_context
def fastq(ctx, indir, spring_path, first, second, dry_run):
    """Run whole pipeline by compressing, comparing and deleting original fastqs.
    Either all fastq pairs below a directory or a given pair.
    """
    if dry_run:
        LOG.info("Dry Run! No files will be created or deleted")

    if indir:
        LOG.info(
            "This will recursively compress and delete fastqs in %s",
            indir,
        )
        indir = pathlib.Path(indir)
        if not indir.is_dir():
            LOG.warning("Please specify a directory")
            return
        pairs = find_fastq_pairs(indir)

    else:
        if not (first and second and spring_path):
            LOG.warning("Please specify either a directory or two fastqs and a spring path")
            return
        LOG.info(
            "Running auto, this will compress and delete fastqs %s, %s into %s",
            first,
            second,
            spring_path,
        )

        pairs = [(pathlib.Path(first), pathlib.Path(second), pathlib.Path(spring_path))]

    for pair in pairs:
        first_fastq = pair[0]
        second_fastq = pair[1]
        spring_path = pair[2]
        try:
            ctx.invoke(
                compress_fastq_cmd,
                first_read=str(first_fastq),
                second_read=str(second_fastq),
                spring_path=str(spring_path),
                dry_run=dry_run,
                check_integrity=True,
            )
        except click.Abort:
            LOG.warning("Skip current and continue auto")
            continue

        LOG.info("Deleting original fastqs")
        if not dry_run:
            first_fastq.unlink()
            LOG.info("%s deleted", first_fastq)
            second_fastq.unlink()
            LOG.info("%s deleted", second_fastq)

    LOG.info("crunchy auto fastq completed")


auto.add_command(fastq)



"""Tests for decompress command"""

import pathlib
import shutil
from pathlib import Path

from click.testing import CliRunner

from crunchy.cli.decompress_cmd import cram, decompress, spring
from crunchy.files import fastq_outpaths
from crunchy.integrity import get_checksum


def nr_files(dirpath: pathlib.Path) -> int:
    """Count the number of files in a directory"""
    nr_files_indir = 0
    for path in dirpath.iterdir():
        if not path.is_file():
            continue
        nr_files_indir += 1
    return nr_files_indir


def test_decompress_cram(base_context, cram_tmp_file, bam_tmp_path):
    """Test to run the decompress cram command"""
    # GIVEN a cli runner
    runner = CliRunner()
    # WHEN running the decompress cram command
    result = runner.invoke(cram, [str(cram_tmp_file), "-b", str(bam_tmp_path)], obj=base_context)
    # THEN assert the command succeeds
    assert result.exit_code == 0


def test_decompress_cmd():
    """Test to run the decompress base command"""
    # GIVEN a cli runner
    runner = CliRunner()
    # WHEN running the compress command with dry_run
    result = runner.invoke(decompress, obj={})
    # THEN assert the command was succesful even without a valid api
    assert result.exit_code == 0


def test_decompress_spring_non_existing_spring(base_context, spring_tmp_path):
    """Test to run decompress spring with a non existing spring file"""
    # GIVEN a cli runner
    runner = CliRunner()
    # GIVEN non existing spring file
    assert not spring_tmp_path.exists()
    # WHEN running the de compress command with non existing spring file
    result = runner.invoke(spring, [str(spring_tmp_path)], obj=base_context)
    # THEN assert the command fails since the spring file needs to exist
    assert result.exit_code == 2


def test_decompress_spring_no_fastq(spring_tmp_file, base_context):
    """Test to run decompress spring command without specifying the out paths"""
    # GIVEN a cli runner
    runner = CliRunner()
    # GIVEN two non existing fastq paths
    fastq1, fastq2 = fastq_outpaths(spring_tmp_file)
    assert not (fastq1.exists() or fastq2.exists())
    # WHEN running the decompress command without fastq files
    result = runner.invoke(spring, [str(spring_tmp_file)], obj=base_context)
    # THEN assert the command was succesful since crunchy creates the file names
    assert result.exit_code == 0


def test_decompress_spring_no_fastq_existing_paths(
    spring_tmp_file, first_read, second_read, base_context
):
    """Test to run the compress base command"""
    # GIVEN a cli runner and the path to an existing spring file
    runner = CliRunner()
    # GIVEN two existing fastq files
    fastqs = fastq_outpaths(spring_tmp_file)
    first = fastqs[0]
    second = fastqs[1]
    shutil.copy(str(first_read), str(first))
    shutil.copy(str(second_read), str(second))
    assert first.exists()
    # WHEN running the decompress command without fastq files
    result = runner.invoke(spring, [str(spring_tmp_file)], obj=base_context)
    # THEN assert the command fails since the autogenerated fastq files exists
    assert result.exit_code == 1


def test_decompress_spring_with_fastq(
    spring_tmp_file, first_tmp_path, second_tmp_path: Path, base_context
):
    """Test to run the compress base command"""
    # GIVEN a cli runner
    runner = CliRunner()
    # WHEN running the decompress command with fastq files
    result = runner.invoke(
        spring,
        [str(spring_tmp_file), "-f", str(first_tmp_path), "-s", second_tmp_path.as_posix()],
        obj=base_context,
    )
    # THEN assert the command was succesful
    assert result.exit_code == 0


def test_decompress_spring_with_result(
    spring_tmp_file,
    first_tmp_path,
    second_tmp_path: Path,
    base_context,
    first_read,
    second_read,
):
    """Test to run decompress spring command where the mock creates output files"""
    # GIVEN a cli runner
    runner = CliRunner()
    # GIVEN fasq paths that does not exist
    assert not (first_tmp_path.exists() or second_tmp_path.exists())
    # GIVEN a mock that produces output
    base_context["spring_api"]._create_output = True
    base_context["spring_api"]._fastq1 = first_read
    base_context["spring_api"]._fastq2 = second_read
    # WHEN running the decompress command with fastq files
    result = runner.invoke(
        spring,
        [str(spring_tmp_file), "-f", str(first_tmp_path), "-s", second_tmp_path.as_posix()],
        obj=base_context,
    )
    # THEN assert the command was succesful
    assert result.exit_code == 0
    # THEN assert that the fastq files where created by the mock
    assert first_tmp_path.exists() and second_tmp_path.exists()


def test_decompress_spring_with_fastq_and_integrity_check(
    spring_tmp_file,
    first_tmp_path,
    second_tmp_path: Path,
    base_context,
    first_read,
    second_read,
):
    """Test to run decompress spring command with integrity check."""
    # GIVEN a cli runner
    runner = CliRunner()
    # GIVEN fasq paths that does not exist
    assert not (first_tmp_path.exists() or second_tmp_path.exists())
    # GIVEN a mock that produces output
    base_context["spring_api"]._create_output = True
    base_context["spring_api"]._fastq1 = first_read
    base_context["spring_api"]._fastq2 = second_read
    # GIVEN checksums for the original fastq files
    checksum1 = get_checksum(first_read)
    checksum2 = get_checksum(second_read)
    # WHEN running the decompress command with fastq files
    result = runner.invoke(
        spring,
        [
            str(spring_tmp_file),
            "-f",
            str(first_tmp_path),
            "-s",
            second_tmp_path.as_posix(),
            "--first-checksum",
            checksum1,
            "--second-checksum",
            checksum2,
        ],
        obj=base_context,
    )
    # THEN assert the command was succesful
    assert result.exit_code == 0
    # THEN assert that the integrity check was successful
    assert first_tmp_path.exists() and second_tmp_path.exists()


def test_decompress_spring_with_fastq_failing_integrity_check(
    spring_tmp_file,
    first_tmp_path,
    second_tmp_path: Path,
    base_context,
    first_read,
    second_read,
):
    """Test to run decompress spring command with a failing integrity check"""
    # GIVEN a cli runner
    runner = CliRunner()
    # GIVEN fasq paths that does not exist
    assert not (first_tmp_path.exists() or second_tmp_path.exists())
    # GIVEN a mock that produces output
    base_context["spring_api"]._create_output = True
    base_context["spring_api"]._fastq1 = first_read
    base_context["spring_api"]._fastq2 = second_read
    # GIVEN checksums for the original fastq files
    checksum1 = get_checksum(first_read)
    # GIVEN wrong checksum for second read
    checksum2 = get_checksum(first_read)
    # WHEN running the decompress command with fastq files
    result = runner.invoke(
        spring,
        [
            str(spring_tmp_file),
            "-f",
            str(first_tmp_path),
            "-s",
            second_tmp_path.as_posix(),
            "--first-checksum",
            checksum1,
            "--second-checksum",
            checksum2,
        ],
        obj=base_context,
    )
    # THEN assert the command was succesful
    assert result.exit_code == 1
    # THEN assert that the fastq files are deleted since check failed
    assert not (first_tmp_path.exists() or second_tmp_path.exists())



"""Utilities for crunchy"""

import logging
import pathlib
from typing import Iterator

LOG = logging.getLogger(__name__)


def find_fastq_pairs(directory: pathlib.Path) -> Iterator[tuple]:
    """Loop over all subdirectories and return all fastq pairs found

    The function will assume the naming convention used by illumina and described in detail:
    (https://support.illumina.com/help/BaseSpace_OLH_009008/Content/Source/Informatics/BS/Naming\
     Convention_FASTQ-files-swBS.htm)

    Search for all files that have fq or fastq in its name. If found check if they are read pairs.
    Then check if both reads in pair is available.

    yields:
        tuples where first two elements are pairs and third is spring name
    """
    LOG.info("Find all pairs in %s", directory)
    valid_fastq_endings = set([".fq", ".fastq"])
    reads_found = set()
    for pth in directory.rglob("*"):
        endings = pth.suffixes
        if not set(endings).intersection(valid_fastq_endings):
            continue

        file_name = pth.name
        file_parent = pth.parent

        splitted = file_name.split("_")
        if pth in reads_found:
            LOG.debug("Read already found: %s", pth)
            continue
        if len(splitted) < 3:
            LOG.info("Fastq filename %s does not follow illumina conventions", file_name)
            continue
        # Check if we have a part of a read pair
        spring_name = pathlib.Path("_".join(splitted[:-2]))
        if splitted[-2] in ["R1", "R2"]:
            pair_str = str(spring_name) + "_{}_" + splitted[-1]
            read_1 = pathlib.Path(file_parent, pair_str.format("R1"))
            if not read_1.exists():
                LOG.warning("Could not find first read in pair: %s", read_1)
                continue
            read_2 = pathlib.Path(file_parent, pair_str.format("R2"))
            if not read_2.exists():
                LOG.warning("Could not find second read in pair: %s", read_2)
                continue

            reads_found.add(read_1)
            reads_found.add(read_2)

            fastq_info = (
                read_1,
                read_2,
                pathlib.Path(file_parent).joinpath(spring_name).with_suffix(".spring"),
            )

            yield fastq_info



"""Test the __main__ module"""

import subprocess


def test_main():
    """Test to run the test module"""
    completed = subprocess.run(["python", "-m", "crunchy"])
    assert completed.returncode == 0



"""Test for the auto functionality in crunchy"""

import pathlib

from click.testing import CliRunner

from crunchy.cli.auto_cmd import auto, fastq


def nr_files(dirpath: pathlib.Path) -> int:
    """Count the number of files in a directory"""
    nr_files_indir = 0
    for path in dirpath.iterdir():
        if not path.is_file():
            continue
        nr_files_indir += 1
    return nr_files_indir


def test_auto_base():
    """Test base command"""
    # GIVEN a cli runner
    runner = CliRunner()
    # WHEN running the auto base command
    result = runner.invoke(auto)
    # THEN assert it exits without problems
    assert result.exit_code == 0


def test_auto_pairs(base_context, first_tmp_file, second_tmp_file):
    """Test auto fastq with files"""
    # GIVEN a cli runner
    runner = CliRunner()
    # WHEN running the auto base command
    result = runner.invoke(
        fastq, ["-f", first_tmp_file, "-s", second_tmp_file, "--yes"], obj=base_context
    )
    # THEN assert it exits without problems
    assert result.exit_code == 0


def test_auto_pairs_and_spring(base_context, first_tmp_file, second_tmp_file, spring_tmp_path):
    """Test auto fastq with files"""
    # GIVEN a cli runner
    runner = CliRunner()
    # WHEN running the auto base command
    result = runner.invoke(
        fastq, ["-f", first_tmp_file, "-s", second_tmp_file, "--yes"], obj=base_context
    )
    # THEN assert it exits without problems
    assert result.exit_code == 0


def test_auto_dir_dry_run(base_context, project_dir):
    """Test auto fastq with dir"""
    # GIVEN a cli runner
    runner = CliRunner()
    # WHEN running the auto base command
    result = runner.invoke(
        fastq, ["--indir", str(project_dir), "--yes", "--dry-run"], obj=base_context
    )
    # THEN assert it exits without problems
    assert result.exit_code == 0


def test_auto_dir_nodir_dry_run(base_context, first_read):
    """Test auto fastq when dir is no dir"""
    # GIVEN a cli runner
    runner = CliRunner()
    # WHEN running the auto base command
    result = runner.invoke(
        fastq, ["--indir", str(first_read), "--yes", "--dry-run"], obj=base_context
    )
    # THEN assert it exits without running
    assert result.exit_code == 0



"""
Code to handle communications to the shell.
"""

import copy
import logging
import subprocess
from pathlib import Path
from subprocess import CalledProcessError
from typing import Generator, List, Optional

LOG = logging.getLogger(__name__)


class Process:
    """Class to handle communication with other programs via the shell.

    The other parts of the code should not need to have any knowledge about how the processes are
    called, that will be handled in this module.Output form stdout and stdin will be handeld here.
    """

    def __init__(
        self, binary: str, config: Optional[str] = None, config_parameter: str = "--config"
    ):
        """
        Args:
            binary(str): Path to binary for the process to use
            config(str): Path to config if used by process
        """
        super(Process, self).__init__()
        self.binary: str = binary
        LOG.info(f"Initialising Process with binary: {self.binary}")
        self.base_call: List[str] = [self.binary]
        if config:
            self.base_call.extend([config_parameter, config])
        LOG.info(f"Use base call {self.base_call}")
        self._stdout: str = ""
        self._stderr: str = ""

    def run_command(self, parameters=None):
        """Execute a command in the shell.

        Args:
            parameters(list)
        """
        command = copy.deepcopy(self.base_call)
        if parameters:
            command.extend(parameters)

        LOG.info("Running command %s", " ".join(command))
        res = subprocess.run(command, check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        self.stdout = res.stdout.decode("utf-8").rstrip()
        self.stderr = res.stderr.decode("utf-8").rstrip()
        if res.returncode != 0:
            LOG.critical(f"Call {command} exit with a non zero exit code")
            LOG.critical(self.stderr)
            raise CalledProcessError(command, res.returncode)

        return res.returncode

    @property
    def stdout(self):
        """Fetch stdout."""
        return self._stdout

    @stdout.setter
    def stdout(self, text):
        self._stdout = text

    @stdout.deleter
    def stdout(self):
        del self._stdout

    @property
    def stderr(self):
        """Fetch stderr."""
        return self._stderr

    @stderr.setter
    def stderr(self, text):
        self._stderr = text

    @stderr.deleter
    def stderr(self):
        del self._stderr

    def stdout_lines(self) -> Generator[List[str], None, None]:
        """Iterate over the lines in self.stdout."""
        yield from self.stdout.split("\n")

    def stderr_lines(self) -> Generator[List[str], None, None]:
        """Iterate over the lines in self.stderr."""
        yield from self.stderr.split("\n")

    def __repr__(self) -> str:
        return f"Process:base_call:{self.base_call}"


class SpringProcess(Process):
    """Process to deal with String commands."""

    def __init__(self, binary: str, threads: int = 8, tmp_dir: Optional[str] = None):
        """Initialise a spring process."""
        super().__init__(binary)
        self.threads: int = threads
        self.tmp: Optional[str] = tmp_dir

    def decompress(self, spring_path: Path, first: Path, second: Path) -> bool:
        """Run the spring decompress command."""
        parameters = ["-d", "-i", str(spring_path), "-o", str(first), str(second)]
        if first.suffix == ".gz":
            LOG.info("Compressing to gzipped format")
            parameters.append("-g")

        if self.tmp:
            parameters.extend(["--working-dir", self.tmp])

        LOG.info("Decompressing Spring compressed file")
        self.run_command(parameters)
        success = False
        time_to_decompress = "unknown"
        for line in self.stdout_lines():
            line = line.lower()
            if "decompression done" in line:
                success = True
            if "total time for decompression" in line:
                time_to_decompress = line.split(" ")[-2]

        if success:
            LOG.info("Spring decompression successfully completed!")
            LOG.info(f"Time to decompress: {time_to_decompress}")
            return True

        LOG.error("Spring decompression failed")
        LOG.error(self.stderr)
        return False

    def compress(self, first: Path, second: Path, outfile: Path) -> bool:
        """Run the spring compression command."""
        parameters = [
            "-c",
            "-i",
            str(first),
            str(second),
            "-o",
            str(outfile),
            "-t",
            str(self.threads),
        ]

        if first.suffix == ".gz":
            LOG.info("File(s) are gzipped")
            parameters.append("-g")

        if self.tmp:
            parameters.extend(["--working-dir", self.tmp])

        LOG.info("Compressing FASTQ to Spring")
        self.run_command(parameters)
        success = False
        time_to_compress = "unknown"
        for line in self.stdout_lines():
            line = line.lower()
            if "compression done" in line:
                success = True
            if "total time for compression" in line:
                time_to_compress = line.split(" ")[-2]

        if success:
            LOG.info("Spring compression successfully completed!")
            LOG.info(f"Time to compress: {time_to_compress}")
            return True

        LOG.error("Spring compression failed")
        LOG.error(self.stderr)
        return False

    def __repr__(self):
        return f"SpringProcess:base_call:{self.base_call}"


class CramProcess(Process):
    """Process to deal with CRAM commands."""

    def __init__(self, binary: str, refgenome_path: str, threads: int = 8):
        """Initialise a spring process."""
        super().__init__(binary)
        self.refgenome_path: str = refgenome_path
        self.threads: int = threads

    def decompress(self, cram_path: Path, bam_path: Path) -> bool:
        """Convert CRAM to BAM."""
        LOG.info(f"Decompressing cram {cram_path} to bam {bam_path}")
        parameters = [
            "view",
            "-b",
            "-o",
            str(bam_path),
            "-T",
            self.refgenome_path,
            str(cram_path),
        ]
        self.run_command(parameters)
        return True

    def compress(self, bam_path: Path, cram_path: Path) -> bool:
        """Convert BAM to CRAM."""
        LOG.info(f"Compressing BAN {bam_path} to CRAM {cram_path}")
        parameters = [
            "view",
            "-C",
            "-T",
            self.refgenome_path,
            str(bam_path),
            "-o",
            str(cram_path),
        ]
        self.run_command(parameters)
        self.index(cram_path)
        return True

    @staticmethod
    def get_index_path(file_path: Path) -> Path:
        """Create a index path based on a file name."""
        index_suffix = ".bai" if file_path.suffix == ".bam" else ".crai"
        return file_path.with_suffix(file_path.suffix + index_suffix)

    def index(self, file_path: Path):
        """Index a BAM or CRAM file."""
        LOG.info(f"Creating index for {file_path}")
        index_path = self.get_index_path(file_path)
        parameters = ["index", str(file_path), str(index_path)]
        self.run_command(parameters)

    def self_check(self):
        """Run a check and see that all parameters are valid."""
        LOG.info("Check that Cram process is correctly initialized.")
        if self.refgenome_path is None:
            LOG.warning("Please specify the path to a reference genome")
            raise SyntaxError
        if not Path(self.refgenome_path).exists():
            LOG.warning("Reference genome %s does not exist", self.refgenome_path)
            raise FileNotFoundError

    def __repr__(self):
        return f"CramProcess:base_call:{self.base_call}, refgenome_path:{self.refgenome_path}"



"""Tests for the command module"""


def test_get_index_path_cram(cram_api, cram_tmp_path):
    """test to create a index path"""
    # GIVEN a cram path and a cram api
    assert cram_tmp_path.suffix == ".cram"
    # WHEN creating a index
    index = cram_api.get_index_path(cram_tmp_path)
    # THEN assert that the index has the correct suffix
    assert index.suffix == ".crai"
    assert set(index.suffixes) == set([".cram", ".crai"])


def test_get_index_path_bam(cram_api, bam_tmp_path):
    """test to create a index path"""
    # GIVEN a bam path and a cram api
    assert bam_tmp_path.suffix == ".bam"
    # WHEN creating a index
    index = cram_api.get_index_path(bam_tmp_path)
    # THEN assert that the index has the correct suffix
    assert index.suffix == ".bai"
    assert set(index.suffixes) == set([".bam", ".bai"])



"""Fixtures to test the command module"""

import pytest


@pytest.fixture
def spring_api(real_spring_api):
    """Return a spring api that runs spring"""
    return real_spring_api


@pytest.fixture
def cram_api(real_cram_api):
    """Return a cram api"""
    return real_cram_api



"""Tests for integrity module"""

import hashlib

from crunchy.integrity import get_checksum


def test_generate_md5(dummy_file_path):
    """Test to generate a md5"""
    # GIVEN a file and a calculated md5
    with open(dummy_file_path, "rb") as infile:
        content = infile.read()
    md5 = hashlib.md5(content).hexdigest()

    # WHEN generating a md5 from that file
    res = get_checksum(dummy_file_path, "md5")

    # THEN assert a md5 was created and returned as a string
    assert isinstance(res, str)
    # THEN the md5 is correct
    assert res == md5


def test_generate_sha1(dummy_file_path):
    """Test to generate a sha1 checksum"""
    # GIVEN a file and a calculated sha1
    with open(dummy_file_path, "rb") as infile:
        content = infile.read()
    sha1 = hashlib.sha1(content).hexdigest()

    # WHEN generating a sha1 from that file
    res = get_checksum(dummy_file_path, "sha1")

    # THEN assert a sha1 was created and returned as a string
    assert isinstance(res, str)
    # THEN the sha1 is correct
    assert res == sha1


def test_compare_checksums(dummy_file_path):
    """Test to compare two checksums"""
    # GIVEN a file

    # WHEN generating a checksum from that file
    res = get_checksum(dummy_file_path)
    res_2 = get_checksum(dummy_file_path)
    # THEN assert the checksums are correct
    assert res == res_2


def test_generate_checksum_read_1(first_read, checksum_first_read):
    """Test to generate checksum for a fastq file"""
    # GIVEN a fastq file and the corresponding sha256 checksum

    # WHEN generating a checksum for that fastq file
    res = get_checksum(first_read)

    # THEN assert that the checksums are the same
    assert res == checksum_first_read



"""CLI functionality for crunchy"""

import logging

import click
import coloredlogs

from crunchy.command import CramProcess, SpringProcess
from crunchy.version import __version__

from .auto_cmd import auto
from .checksum_cmd import checksum
from .compare_cmd import compare
from .compress_cmd import compress
from .decompress_cmd import decompress

LOG = logging.getLogger(__name__)
LOG_LEVELS = ["DEBUG", "INFO", "WARNING"]


@click.group()
@click.option("--spring-binary", default="spring", show_default=True, help="Path to spring binary")
@click.option(
    "--samtools-binary",
    default="samtools",
    show_default=True,
    help="Path to spring binary",
)
@click.option(
    "--threads",
    "-t",
    default=8,
    show_default=True,
    help="Number of threads to use for spring compression",
)
@click.version_option(__version__)
@click.option(
    "--reference",
    "-r",
    help="Path to reference genome",
)
@click.option(
    "--log-level",
    default="INFO",
    type=click.Choice(LOG_LEVELS),
    help="Choose what log messages to show",
)
@click.option(
    "--tmp-dir",
    help="If specific temp dir should be used",
)
@click.pass_context
def base_command(ctx, spring_binary, samtools_binary, threads, reference, log_level, tmp_dir):
    """Base command for crunchy

    \b
                  .---. .---.
                 :     : o   :    me want cookie!
             _..-:   o :     :-.._    /
         .-''  '  `---' `---' "   ``-.
       .'   "   '  "  .    "  . '  "  `.
      :   '.---.,,.,...,.,.,.,..---.  ' ;
      `. " `.                     .' " .'
       `.  '`.                   .' ' .'
        `.    `-._           _.-' "  .'  .----.
          `. "    '"--...--"'  . ' .'  .'  o   `.
          .'`-._'    " .     " _.-'`. :       o  :
        .'      ```--.....--'''    ' `:_ o       :
      .'    "     '         "     "   ; `.;";";";'
     ;         '       "       '     . ; .' ; ; ;
    ;     '         '       '   "    .'      .-'
    '  "     "   '      "           "    _.-'
    """
    coloredlogs.install(level=log_level)
    spring_api = SpringProcess(spring_binary, threads, tmp_dir)
    ctx.obj = {}
    ctx.obj["spring_api"] = spring_api
    cram_api = CramProcess(samtools_binary, reference, threads)
    ctx.obj["cram_api"] = cram_api
    LOG.info("Running crunchy")


base_command.add_command(compare)
base_command.add_command(decompress)
base_command.add_command(compress)
base_command.add_command(auto)
base_command.add_command(checksum)



"""Code for checksum module."""

import logging
import pathlib

import click

from crunchy.integrity import compare_elements, get_checksum

LOG = logging.getLogger(__name__)


@click.command()
@click.option("--first", "-f", type=click.Path(exists=True), required=True)
@click.option("--second", "-s", type=click.Path(exists=True))
@click.option(
    "--checksum",
    "-c",
    help="If the file should be compared to a checksum directly",
)
@click.option("--algorithm", "-a", type=click.Choice(["md5", "sha1", "sha256"]), default="sha256")
@click.option("--dry-run", is_flag=True)
def compare(first, second, algorithm, checksum, dry_run):
    """Compare two files by generating checksums. Fails if two files differ.

    Either the checksum of two files can be compared. Files will be decompressed and checksums
    calculated. Or the checksum of a file can be compared to a checksum string given on the command
    line. Use --first and --checksum if a file should be compared directly to a checksum.
    """
    LOG.info("Running checksum")
    if second and checksum:
        LOG.error("Use --first only in combination with --checksum")
        raise click.Abort

    if dry_run:
        LOG.info("Dry Run!")

    checksums = []
    if checksum:
        checksums.append(checksum)

    for _infile in [first, second]:
        if not _infile:
            continue
        if dry_run:
            checksums.append("dummy_checksum")
            continue
        _infile = pathlib.Path(_infile)
        checksums.append(get_checksum(_infile, algorithm))

    if not dry_run and not compare_elements(checksums):
        LOG.warning(f"Checksums for {first} and {second} are NOT the same")
        raise click.Abort

    LOG.info(f"Checksum: {checksums[0]}")
    LOG.info("All checksums are the same")



"""Functions to decompress files."""

import logging
from pathlib import Path

from .command import CramProcess, SpringProcess

LOG = logging.getLogger(__name__)


def decompress_spring(
    spring_path: Path,
    first: Path,
    second: Path,
    spring_api: SpringProcess,
    dry_run: bool = False,
) -> bool:
    """Decompress a spring file into two fastq files."""
    spring_path = spring_path.absolute()
    first = first.absolute()
    second = second.absolute()
    LOG.info(f"Decompressing {spring_path} to {first} and {second}")
    if dry_run:
        return True
    return spring_api.decompress(spring_path=spring_path, first=first, second=second)


def decompress_cram(
    cram_path: Path,
    bam_path: Path,
    cram_api: CramProcess,
    dry_run: bool = False,
) -> bool:
    """Decompress a cram file into a bam file."""
    cram_path = cram_path.absolute()
    bam_path = bam_path.absolute()
    LOG.info(f"Decompressing {cram_path} to {bam_path}")
    if dry_run:
        return True
    return cram_api.decompress(cram_path=cram_path, bam_path=bam_path)



"""Tests for compression."""

import pathlib
import tempfile

from crunchy.command import CramProcess, SpringProcess
from crunchy.compress import compress_cram, compress_spring


def test_compress_spring(first_read, second_read, spring_api: SpringProcess):
    """Test the compress function."""
    # GIVEN two files with reads from read pair, a spring api and a outfile
    outpath = pathlib.Path(tempfile.NamedTemporaryFile().name)
    spring_api.compress_success = True
    # WHEN running the compression
    res = compress_spring(
        first_read=first_read,
        second_read=second_read,
        outfile=outpath,
        spring_api=spring_api,
    )
    # THEN assert that the run was successful
    assert res is True


def test_compress_spring_dry_run(first_read, second_read, spring_api: SpringProcess):
    """Test the compress function."""
    # GIVEN two files with reads from read pair, a spring api and a outfile
    outpath = pathlib.Path(tempfile.NamedTemporaryFile().name)
    # WHEN running the compression
    res = compress_spring(
        first_read=first_read,
        second_read=second_read,
        outfile=outpath,
        spring_api=spring_api,
        dry_run=True,
    )
    # THEN assert that the run was successful
    assert res is True


def test_compress_cram(bam_path, cram_path, cram_api):
    """Test the compress cram function"""
    # GIVEN a bam_path, a cram path and a cram_api
    # WHEN running the compression
    res = compress_cram(bam_path=bam_path, cram_path=cram_path, cram_api=cram_api)
    # THEN assert that the run was successful
    assert res is True


def test_compress_cram_dry_run(bam_path, cram_api: CramProcess, cram_path):
    """Test the compress cram function"""
    # GIVEN a bam_path, a cram path

    # WHEN running the compression
    res = compress_cram(bam_path=bam_path, cram_path=cram_path, cram_api=cram_api, dry_run=True)
    # THEN assert that the run was successful
    assert res is True



"""Code for decompress CLI fuctions"""

import logging
import pathlib

import click

from crunchy.cli.compare_cmd import compare
from crunchy.decompress import decompress_cram, decompress_spring
from crunchy.files import fastq_outpaths

LOG = logging.getLogger(__name__)


@click.group()
def decompress():
    """Decompress genomic files"""
    LOG.info("Running decompress")


@click.command()
@click.argument("spring-path", type=click.Path(exists=True))
@click.option(
    "--first-read",
    "-f",
    type=click.Path(exists=False),
    help="Fastq file that spring archive will be decompressed to. First read in pair",
)
@click.option(
    "--second-read",
    "-s",
    type=click.Path(exists=False),
    help="Fastq file that spring archive will be decompressed to. Second read in pair",
)
@click.option(
    "--first-checksum",
    help="Checksum from the original fastq file, first in pair",
)
@click.option(
    "--second-checksum",
    help="Checksum from the original fastq file, second in pair",
)
@click.option(
    "--dry-run",
    is_flag=True,
    help="Skip deleting original files",
)
@click.pass_context
def spring(ctx, spring_path, first_read, second_read, first_checksum, second_checksum, dry_run):
    """Decompress a spring file to fastq files"""
    LOG.info("Running decompress spring")
    spring_api = ctx.obj.get("spring_api")
    spring_path = pathlib.Path(spring_path)

    if not (first_read or second_read):
        LOG.warning("No filenames provided. Guess fastq file names")
        fastqs = fastq_outpaths(spring_path)
        first_read = fastqs[0]
        second_read = fastqs[1]

    first_read = pathlib.Path(first_read)
    second_read = pathlib.Path(second_read)
    if first_read.exists() or second_read.exists():
        LOG.error("Outpath(s) already exists! Specify new with '-f', '-s'")
        raise click.Abort()

    decompress_spring(
        spring_path=spring_path,
        first=first_read,
        second=second_read,
        spring_api=spring_api,
        dry_run=dry_run,
    )

    if not (first_checksum and second_checksum):
        LOG.info("Spring file decompressed")
        return

    try:
        ctx.invoke(compare, first=str(first_read), checksum=first_checksum, dry_run=dry_run)
        ctx.invoke(compare, first=str(second_read), checksum=second_checksum, dry_run=dry_run)
    except click.Abort:
        LOG.error("Uncompressed spring differ from given checksum")
        LOG.info("Deleting decompressed fastq files")
        first_read.unlink()
        second_read.unlink()
        raise click.Abort

    LOG.info("Spring file decompressed and verified")


@click.command()
@click.argument("cram-path", type=click.Path(exists=True))
@click.option(
    "--bam-path",
    "-b",
    type=click.Path(exists=False),
    required=True,
    help="Path to bam file",
)
@click.option(
    "--dry-run",
    is_flag=True,
    help="Skip deleting original files",
)
@click.pass_context
def cram(ctx, cram_path, bam_path, dry_run):
    """Decompress a cram file to bam file"""
    LOG.info("Running decompress cram")
    cram_api = ctx.obj.get("cram_api")
    cram_path = pathlib.Path(cram_path)
    bam_path = pathlib.Path(bam_path)

    decompress_cram(
        cram_path=cram_path,
        bam_path=bam_path,
        cram_api=cram_api,
        dry_run=dry_run,
    )


decompress.add_command(spring)
decompress.add_command(cram)
